////////////////////////////////////////////////////////////////////// 
// libsrc/Pedigree.h 
// (c) 2000-2005 Goncalo Abecasis
// 
// This file is distributed as part of the PEDSTATS source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile PEDSTATS.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Thursday August 11, 2005
// 
 
#ifndef _PEDIGREE_H_
#define _PEDIGREE_H_

#include "Constant.h"

#include <stdio.h>

#include "PedigreeAlleles.h"
#include "PedigreePerson.h"
#include "PedigreeGlobals.h"
#include "PedigreeFamily.h"
#include "PedigreeDescription.h"
#include "PedigreeAlleleFreq.h"

class Pedigree : public PedigreeGlobals
   {
   public:
      static int           writeHalves;
      static String        missing;

      PedigreeDescription  pd;
      int                  size;
      int                  count;
      Person **            persons;
      int                  familyCount;
      Family **            families;
      int                  haveTwins;

   Pedigree();
   virtual ~Pedigree();

   void Prepare(FILE * input);      // Read pedigree parameters from data file
   void Load(FILE * input);         // Read pedigree from pedigree file
   void LoadMendel(FILE * input);   // Read pedigree in Mendel format
   void Prepare(const char * input);// Read pedigree parameters from named file
   void Load(const char * input);   // Read pedigree parameters from named file

   void PrepareDichotomization();   // Register dummy affections for each trait
   int  Dichotomize(int trait, double mean = _NAN_);
   void DichotomizeAll(double mean = _NAN_);

   void WriteDataFile(FILE * output);           // Write data file
   void WritePedigreeFile(FILE * output);       // Write pedigree file
   void WriteDataFile(const char * output);     // Write named data file
   void WritePedigreeFile(const char * output); // Write named pedigree file
   void WritePerson(FILE * output, int who,     // Write a single person
        const char * famid = NULL,              // if supplied, famid, pid,
        const char * pid = NULL,                // fatid and motid allow a
        const char * fatid = NULL,              // pedigree or person to
        const char * motid = NULL);             // be renamed / restructured
   void WriteRecodedPerson(                     // Like write person, but uses
        FILE * output, int who,                 // user supplied markerInfo
        MarkerInfo ** markerInfo,               // array to recode marker
        const char * famid = NULL,              // alleles as they are written
        const char * pid = NULL,
        const char * fatid = NULL,
        const char * motid = NULL);

   virtual void Sort();                              // Sorts the pedigree items
   Family * FindFamily(const char * famid);  // Find a family
   Person * FindPerson(const char * famid,   // Find an individual
                       const char * pid);

   // functions dealing with genetic markers
   // Counts the alleles at a marker
   int  CountAlleles(int marker);                     
   // Lumps together rare alleles
   void LumpAlleles(double treshold);               
   // Calculate allele frequencies
   void EstimateFrequencies(int estimator, bool quiet = false); 

   // shorthand operators
   Person & operator [] (int i)
      {
      return *(persons[i]);
      }

   // Perform a basic inheritance check
   bool InheritanceCheck();
   bool AutosomalCheck();
   bool SexLinkedCheck();
   bool TwinCheck();

   // Merge twins into a single individual
   void MergeTwins();

   // Remove individuals with no data from pedigree
   void Trim();
   void ZeroFounderParents();

   // Add all individuals in family with famid = id to new_ped
   void ExtractFamily(int id, Pedigree & new_ped);
   // Add individuals with affection status target_status for affection a to new_ped
   void ExtractOnAffection(int a, Pedigree & new_ped, int target_status = 2);
   
   private:
      void Grow();
      void Add(Person & rhs);

      static int ComparePersons(const Person ** p1, const Person ** p2);
      static int CompareParents(const Person ** p1, const Person ** p2);

      void MakeSibships();
      void MakeFamilies();

      Person * FindPerson(const char * famid, const char * pid, int universe);

      void ShowTrimHeader(bool & flag);
   };

#endif


 
 
 
 
 
 
 
