////////////////////////////////////////////////////////////////////// 
// pedstats/PedstatsHWE.h 
// (c) 2000-2005 Goncalo Abecasis, 2002-2005 Jan Wigginton
// 
// This file is distributed as part of the PEDSTATS source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile PEDSTATS.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Thursday August 11, 2005
// 
 
#ifndef __PEDSTATSHWE_H__
#define __PEDSTATSHWE_H__

#include "StringArray.h"
#include "MathMatrix.h"
#include "FilteredPedigree.h"
#include "IntArray.h"

class PedstatsHWE
   {
   public:

      bool foundersOnly;
      double significanceCutoff;

      PedstatsHWE(FilteredPedigree & p);
      ~PedstatsHWE() {}

      // Index of current marker with respect to Pedigree globals
      int currMarker;

      // Number of alleles at current marker
      int  alleleCount;

      // Number of alleles after pooling
      int  realAlleles;

      // Number of allele that has pooled alleles added to counts
      int  poolAllele;

      // Names of pooled alleles
      String pooledNames;

      // Names for all alleles
      StringArray alleleNames;

      // Allele frequencies after pooling
      Vector      alleleFreqs;

      // Indices in allele freq array corresponding to rare alleles
      IntArray    pooledAlleles;

      // If current marker is tested with the exact test
      bool useExact;

      // If current marker was not tested due to error conditions such
      // as low allele counts, only one allele, etc
      bool testPerformed;

      // P-value, chi-square from test of current marker
      double pValue, chisq;

      // Expected heterozygotes for current marker
      int    expectedHets;

      // Smallest, largest allele for current marker
      int    smallestAllele, largestAllele;

      // Number of failed tests, number of tests done with some alleles pooled for the test
      // Number of failed tests with some alleles pooled
      int failedTests, pooledTests, failedPooledTests;
      
      // Number of failed tests at 0.01 significance cutoff
      int failedTests_01;

      // Number of markers tested, number of attempted tests
      int markerTests, attemptedTests;


      // Total genotypes
      int totalGenotypes;

      // Vector of heterozygote probabilities for the exact test
      Vector hetProbs;

      // Observed and expected genotype counts with respect to pooled alleles
      Matrix realExpected, realObserved;

      // Names of markers that have not been tested due to sparse cells or low allele counts
      StringArray sparseNames, lowAlleleCountNames;

      bool TestMarker(int m, bool founders_only = false);

   private:

      // Alias for pedigree
      FilteredPedigree & ped;

      // Raw observed genotype counts
      Matrix observed;

      double  RunExactTest();
      double  ExactHWE(int rare , int heterozygotes);

      double DetermineGenotypeCounts();
      void   DetermineAlleleNames();
      void   DetermineAlleleRange();
      int    DetermineAlleleFreqs(bool final_count);
      void   DeterminePooledAlleleInfo(int m);
      int    DetermineExpectedHeterozygotes();

      void   PoolAlleles(Vector & freqs);
      void   InitializeArrays();

   };

#endif

 
